""" JWT Authentication class. """

import logging

from django.contrib.auth import get_user_model
from django.middleware.csrf import CsrfViewMiddleware
from edx_django_utils.cache import RequestCache
from edx_django_utils.monitoring import set_custom_attribute
from jwt import exceptions as jwt_exceptions
from rest_framework import exceptions
from rest_framework_jwt.authentication import JSONWebTokenAuthentication

from edx_rest_framework_extensions.auth.jwt.decoder import (
    configured_jwt_decode_handler,
    unsafe_jwt_decode_handler,
)
from edx_rest_framework_extensions.config import ENABLE_SET_REQUEST_USER_FOR_JWT_COOKIE
from edx_rest_framework_extensions.settings import get_setting


logger = logging.getLogger(__name__)


class JwtAuthenticationError(Exception):
    """
    Custom base class for all exceptions
    """


class JwtSessionUserMismatchError(JwtAuthenticationError):
    pass


class CSRFCheck(CsrfViewMiddleware):
    def _reject(self, request, reason):
        # Return the failure reason instead of an HttpResponse
        return reason


class JwtAuthentication(JSONWebTokenAuthentication):
    """
    JSON Web Token based authentication.

    This authentication class is useful for authenticating a JWT using a secret key. Clients should authenticate by
    passing the token key in the "Authorization" HTTP header, prepended with the string `"JWT "`.

    This class relies on the JWT_AUTH being configured for the application as well as JWT_PAYLOAD_USER_ATTRIBUTES
    being configured in the EDX_DRF_EXTENSIONS config.

    At a minimum, the JWT payload must contain a username. If an email address
    is provided in the payload, it will be used to update the retrieved user's
    email address associated with that username.

    Example Header:
        Authorization: JWT eyJhbGciOiJSUzUxMiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiJmYzJiNzIwMTE0YmIwN2I0NjVlODQzYTc0ZWM2ODNlNiIs
        ImFkbWluaXN0cmF0b3IiOmZhbHNlLCJuYW1lIjoiaG9ub3IiLCJleHA.QHDXdo8gDJ5p9uOErTLZtl2HK_61kgLs71VHp6sLx8rIqj2tt9yCfc_0
        JUZpIYMkEd38uf1vj-4HZkzeNBnZZZ3Kdvq7F8ZioREPKNyEVSm2mnzl1v49EthehN9kwfUgFgPXfUh-pCvLDqwCCTdAXMcTJ8qufzEPTYYY54lY
    """

    def get_jwt_claim_attribute_map(self):
        """ Returns a mapping of JWT claims to user model attributes.

        Returns
            dict
        """
        return get_setting('JWT_PAYLOAD_USER_ATTRIBUTE_MAPPING')

    def get_jwt_claim_mergeable_attributes(self):
        """ Returns a list of user model attributes that should be merged into from the JWT.

        Returns
            list
        """
        return get_setting('JWT_PAYLOAD_MERGEABLE_USER_ATTRIBUTES')

    def authenticate(self, request):
        # .. custom_attribute_name: jwt_auth_result
        # .. custom_attribute_description: The result of the JWT authenticate process,
        #      which can having the following values:
        #        'n/a': When JWT Authentication doesn't apply.
        #        'success-auth-header': Successfully authenticated using the Authorization header.
        #        'success-cookie': Successfully authenticated using a JWT cookie.
        #        'forgiven-failure': Returns None instead of failing for JWT cookies. This handles
        #          the case where expired cookies won't prevent another authentication class, like
        #          SessionAuthentication, from having a chance to succeed.
        #          See docs/decisions/0002-remove-use-jwt-cookie-header.rst for details.
        #        'failed-auth-header': JWT Authorization header authentication failed. This prevents
        #          other authentication classes from attempting authentication.
        #        'failed-cookie': JWT cookie authentication failed. This prevents other
        #          authentication classes from attempting authentication.
        #        'user-mismatch-failure': JWT vs session user mismatch found for what would have been
        #          a forgiven-failure, but instead, the JWT failure will be final.
        #        'user-mismatch-enforced-failure': JWT vs session user mismatch found for what would
        #          have been a successful JWT authentication, but we are enforcing a match, and thus
        #          we fail authentication.

        is_authenticating_with_jwt_cookie = self.is_authenticating_with_jwt_cookie(request)
        try:
            user_and_auth = super().authenticate(request)

            # Unauthenticated, CSRF validation not required
            if not user_and_auth:
                set_custom_attribute('jwt_auth_result', 'n/a')
                return user_and_auth

            # Not using JWT cookie, CSRF validation not required
            if not is_authenticating_with_jwt_cookie:
                set_custom_attribute('jwt_auth_result', 'success-auth-header')
                return user_and_auth

            self.enforce_csrf(request)

            # CSRF passed validation with authenticated user

            # adds additional monitoring for mismatches; and raises errors in certain cases
            is_mismatch = self._is_jwt_cookie_and_session_user_mismatch(request)
            if is_mismatch and get_setting(ENABLE_SET_REQUEST_USER_FOR_JWT_COOKIE):
                raise JwtSessionUserMismatchError(
                    'Failing otherwise successful JWT authentication due to session user mismatch '
                    'with set request user.'
                )

            set_custom_attribute('jwt_auth_result', 'success-cookie')
            return user_and_auth

        except JwtSessionUserMismatchError as exception:
            # Warn against these errors because JWT vs session user should not be happening.
            logger.warning('Failed JWT Authentication due to session user mismatch.')
            # .. custom_attribute_name: jwt_auth_failed
            # .. custom_attribute_description: Includes a summary of the JWT failure exception
            #       for debugging.
            set_custom_attribute('jwt_auth_failed', 'Exception:{}'.format(repr(exception)))
            set_custom_attribute('jwt_auth_result', 'user-mismatch-enforced-failure')
            raise

        except Exception as exception:
            # Errors in production do not need to be logged (as they may be noisy),
            # but debug logging can help quickly resolve issues during development.
            logger.debug('Failed JWT Authentication.', exc_info=exception)

            exception_to_report = _deepest_jwt_exception(exception)
            set_custom_attribute('jwt_auth_failed', 'Exception:{}'.format(repr(exception_to_report)))

            if is_authenticating_with_jwt_cookie:
                # This check also adds monitoring details
                is_user_mismatch = self._is_jwt_cookie_and_session_user_mismatch(request)
                if is_user_mismatch:
                    set_custom_attribute('jwt_auth_result', 'user-mismatch-failure')
                    raise
                set_custom_attribute('jwt_auth_result', 'forgiven-failure')
                return None

            set_custom_attribute('jwt_auth_result', 'failed-auth-header')
            raise

    def authenticate_credentials(self, payload):
        """Get or create an active user with the username contained in the payload."""
        # TODO it would be good to refactor this heavily-nested function.
        # pylint: disable=too-many-nested-blocks
        username = self._get_username_from_payload(payload)
        if username is None:
            raise exceptions.AuthenticationFailed('JWT must include a preferred_username or username claim!')
        try:
            user, __ = get_user_model().objects.get_or_create(username=username)
            attributes_updated = False
            attribute_map = self.get_jwt_claim_attribute_map()
            attributes_to_merge = self.get_jwt_claim_mergeable_attributes()
            for claim, attr in attribute_map.items():
                payload_value = payload.get(claim)

                if attr in attributes_to_merge:
                    # Merge new values that aren't already set in the user dictionary
                    if not payload_value:
                        continue

                    current_value = getattr(user, attr, None)

                    if current_value:
                        for (key, value) in payload_value.items():
                            if key in current_value:
                                if current_value[key] != value:
                                    logger.info(
                                        'Updating attribute %s[%s] for user %s with value %s',
                                        attr,
                                        key,
                                        user.id,
                                        value,
                                    )
                                    current_value[key] = value
                                    attributes_updated = True
                            else:
                                logger.info(
                                    'Adding attribute %s[%s] for user %s with value %s',
                                    attr,
                                    key,
                                    user.id,
                                    value,
                                )
                                current_value[key] = value
                                attributes_updated = True
                    else:
                        logger.info('Updating attribute %s for user %s with value %s', attr, user.id, payload_value)
                        setattr(user, attr, payload_value)
                        attributes_updated = True
                else:
                    if getattr(user, attr) != payload_value and payload_value is not None:
                        logger.info('Updating attribute %s for user %s with value %s', attr, user.id, payload_value)
                        setattr(user, attr, payload_value)
                        attributes_updated = True

            if attributes_updated:
                user.save()
        except Exception as authentication_error:
            msg = f'[edx-drf-extensions] User retrieval failed for username {username}.'
            logger.exception(msg)
            raise exceptions.AuthenticationFailed(msg) from authentication_error

        return user

    def enforce_csrf(self, request):
        """
        Enforce CSRF validation for Jwt cookie authentication.

        Copied from SessionAuthentication.
        See https://github.com/encode/django-rest-framework/blob/3f19e66d9f2569895af6e91455e5cf53b8ce5640/rest_framework/authentication.py#L131-L141  # noqa E501 line too long
        """
        check = CSRFCheck(get_response=lambda request: None)
        # populates request.META['CSRF_COOKIE'], which is used in process_view()
        check.process_request(request)
        reason = check.process_view(request, None, (), {})
        if reason:
            # CSRF failed, bail with explicit error message
            raise exceptions.PermissionDenied('CSRF Failed: %s' % reason)

    @classmethod
    def is_authenticating_with_jwt_cookie(cls, request):
        """
        Returns True if authenticating with a JWT cookie, and False otherwise.
        """
        try:
            # If there is a token in the authorization header, it takes precedence in
            # get_token_from_request. This ensures that not only is a JWT cookie found,
            # but that it was actually used for authentication.
            request_token = JSONWebTokenAuthentication.get_token_from_request(request)
            cookie_token = JSONWebTokenAuthentication.get_token_from_cookies(request.COOKIES)
            return cookie_token and (request_token == cookie_token)
        except Exception:  # pylint: disable=broad-exception-caught
            return False

    def _is_jwt_cookie_and_session_user_mismatch(self, request):
        """
        Returns True if JWT cookie and session user do not match, False otherwise.

        Arguments:
            request: The request.

        Other notes:
        - Also adds monitoring details for mismatches.
        - Should only be called for JWT cookies.
        """
        jwt_username, jwt_lms_user_id = self._get_unsafe_jwt_cookie_username_and_lms_user_id(request)

        # add early monitoring for the JWT LMS user_id for observability for a variety of user cases

        # .. custom_attribute_name: jwt_cookie_lms_user_id
        # .. custom_attribute_description: The LMS user_id pulled from the
        #     JWT cookie, or None if the JWT was corrupt and it wasn't found.
        #     Note that the decoding is unsafe, so this isn't just for valid cookies.
        set_custom_attribute('jwt_cookie_lms_user_id', jwt_lms_user_id)

        # If we set the request user in middleware for JWT auth, then we'd actually be checking JWT vs JWT username.
        # Additionally, somehow the setting of request.user and the retrieving of request.user below causes some
        # unknown issue in production-like environments, and this allows us to skip that case.
        if _is_request_user_set_for_jwt_auth():
            return False

        wsgi_request = getattr(request, '_request', request)
        if wsgi_request == request:
            # .. custom_attribute_name: jwt_auth_with_django_request
            # .. custom_attribute_description: There exists custom authentication code in the platform that is
            #      calling JwtAuthentication with a Django request, rather than the expected DRF request. This
            #      custom attribute could be used to track down those usages and find ways to eliminate custom
            #      authentication code that lives outside of this library.
            set_custom_attribute('jwt_auth_with_django_request', True)

        # Get the session-based user from the underlying HttpRequest object.
        # This line taken from DRF SessionAuthentication.
        session_user = getattr(wsgi_request, 'user', None)
        if not session_user:  # pragma: no cover
            return False

        if not session_user.is_authenticated or not session_user.username or session_user.username == jwt_username:
            return False

        # .. custom_attribute_name: jwt_auth_mismatch_session_username
        # .. custom_attribute_description: The session authentication username if it
        #      does not match the JWT cookie username. If there is no session user,
        #      or if it matches the JWT cookie username, this attribute will not be included.
        #      Session authentication may have completed in middleware
        #      before getting to DRF. Although this authentication won't stick,
        #      because it will be replaced by DRF authentication, we record it,
        #      because it sometimes does not match the JWT cookie user.
        set_custom_attribute('jwt_auth_mismatch_session_username', session_user.username)
        # .. custom_attribute_name: jwt_auth_mismatch_jwt_cookie_username
        # .. custom_attribute_description: The JWT cookie username if it
        #      does not match the session authentication username.
        #      See jwt_auth_mismatch_session_username description for more details.
        #      Note that there is a low chance that a corrupt JWT cookie will contain a
        #      username and user id that do not correlate, so we capture the actual username,
        #      even though it is likely redundant to jwt_cookie_lms_user_id. To minimize
        #      the use of PII, this attribute is only captured in the case of a mismatch.
        set_custom_attribute('jwt_auth_mismatch_jwt_cookie_username', jwt_username)

        return True

    def _get_unsafe_jwt_cookie_username_and_lms_user_id(self, request):
        """
        Returns a tuple of the (username, lms user id) from the JWT cookie, or (None, None) if not found.
        """

        # .. custom_attribute_name: jwt_cookie_unsafe_decode_issue
        # .. custom_attribute_description: Since we are doing an unsafe JWT decode, it should generally work unless
        #     the JWT cookie were tampered with. This attribute will contain the value 'missing-claim' if either the
        #     username or user_id claim is missing, or 'decode-error' if the JWT cookie can't be decoded at all. This
        #     attribute will not exist if there is no issue decoding the cookie.

        try:
            cookie_token = JSONWebTokenAuthentication.get_token_from_cookies(request.COOKIES)
            unsafe_decoded_jwt = unsafe_jwt_decode_handler(cookie_token)
            jwt_username = self._get_username_from_payload(unsafe_decoded_jwt)
            jwt_lms_user_id = unsafe_decoded_jwt.get('user_id', None)
            if not jwt_username or not jwt_lms_user_id:
                set_custom_attribute('jwt_cookie_unsafe_decode_issue', 'missing-claim')
        except Exception:  # pylint: disable=broad-exception-caught
            jwt_username = None
            jwt_lms_user_id = None
            set_custom_attribute('jwt_cookie_unsafe_decode_issue', 'decode-error')

        return (jwt_username, jwt_lms_user_id)

    def _get_username_from_payload(self, payload):
        """
        Returns the username from the payload.

        WARNING:
        1. This doesn't play well with JSONWebTokenAuthentication.jwt_get_username_from_payload, but
        some services do not have JWT_PAYLOAD_GET_USERNAME_HANDLER configured.
        2. It's unclear if `username` is used for any old JWTs, but this could probably be removed.
        """
        return payload.get('preferred_username') or payload.get('username')


_IS_REQUEST_USER_SET_FOR_JWT_AUTH_CACHE_KEY = '_is_request_user_for_jwt_set'


def set_flag_is_request_user_set_for_jwt_auth():
    """
    Sets a flag that the shows the request user was set to be based on JWT auth.

    Used to coordinate between middleware and JwtAuthentication. Note that the flag
    is stored in this module to avoid circular dependencies.
    """
    _get_module_request_cache()[_IS_REQUEST_USER_SET_FOR_JWT_AUTH_CACHE_KEY] = True


def is_jwt_authenticated(request):
    successful_authenticator = getattr(request, 'successful_authenticator', None)
    if not isinstance(successful_authenticator, JSONWebTokenAuthentication):
        return False
    if not getattr(request, 'auth', None):
        logger.error(
            'Unexpected error: Used JwtAuthentication, '
            'but the request auth attribute was not populated with the JWT.'
        )
        return False
    return True


def get_decoded_jwt_from_auth(request):
    """
    Grab jwt from request.auth in request if possible.

    Returns a decoded jwt dict if it can be found.
    Returns None if the jwt is not found.
    """
    if not is_jwt_authenticated(request):
        return None

    return configured_jwt_decode_handler(request.auth)


def _deepest_jwt_exception(exception):
    """
    Given an exception, traverse down the __context__ tree
    until you get to the deepest exceptions which is still
    a subclass of PyJWTError.  If no PyJWTError subclass
    exists, then just return the original exception.
    """
    relevant_exception = exception
    cur_exception = exception

    # An exception always has a context but if it's the deepest
    # exception, than __context__ will return None
    while cur_exception.__context__:
        cur_exception = cur_exception.__context__
        if isinstance(cur_exception, jwt_exceptions.PyJWTError):
            relevant_exception = cur_exception

    return relevant_exception


def _get_module_request_cache():
    return RequestCache(__name__).data


def _is_request_user_set_for_jwt_auth():
    """
    Returns whether the request user was set to be based on JWT auth in JwtAuthCookieMiddleware.

    This is a public method to enable coordination with the JwtAuthentication class.
    """
    return _get_module_request_cache().get(_IS_REQUEST_USER_SET_FOR_JWT_AUTH_CACHE_KEY, False)



""" Tests for JWT authentication class. """
from logging import Logger
from unittest import mock

import ddt
from django.contrib.auth import get_user_model
from django.http.cookie import SimpleCookie
from django.test import RequestFactory, TestCase, override_settings
from django.urls import re_path as url_pattern
from django.urls import reverse
from edx_django_utils.cache import RequestCache
from jwt import exceptions as jwt_exceptions
from rest_framework.exceptions import AuthenticationFailed
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework_jwt.authentication import JSONWebTokenAuthentication

from edx_rest_framework_extensions.auth.jwt import authentication
from edx_rest_framework_extensions.auth.jwt.authentication import (
    JwtAuthentication,
    JwtSessionUserMismatchError,
)
from edx_rest_framework_extensions.auth.jwt.cookies import (
    jwt_cookie_header_payload_name,
    jwt_cookie_name,
    jwt_cookie_signature_name,
)
from edx_rest_framework_extensions.auth.jwt.decoder import jwt_decode_handler
from edx_rest_framework_extensions.auth.jwt.tests.utils import (
    generate_jwt_token,
    generate_latest_version_payload,
)
from edx_rest_framework_extensions.config import ENABLE_SET_REQUEST_USER_FOR_JWT_COOKIE
from edx_rest_framework_extensions.tests import factories


User = get_user_model()


class IsAuthenticatedView(APIView):
    authentication_classes = (JwtAuthentication,)
    permission_classes = (IsAuthenticated,)

    def get(self, request):  # pylint: disable=unused-argument
        return Response({'success': True})


urlpatterns = [
    url_pattern(
        r'^isauthenticated/$',
        IsAuthenticatedView.as_view(),
        name='authenticated-view',
    ),
]


@ddt.ddt
class JwtAuthenticationTests(TestCase):
    """ JWT Authentication class tests. """
    def setUp(self):
        super().setUp()
        RequestCache.clear_all_namespaces()

    def get_jwt_payload(self, **additional_claims):
        """ Returns a JWT payload with the necessary claims to create a new user. """
        email = 'gcostanza@gmail.com'
        username = 'gcostanza'
        payload = dict({'preferred_username': username, 'email': email}, **additional_claims)

        return payload

    @ddt.data(True, False)
    def test_authenticate_credentials_user_creation(self, is_staff):
        """ Test whether the user model is being created and assigned fields from the payload. """

        payload = self.get_jwt_payload(administrator=is_staff)
        user = JwtAuthentication().authenticate_credentials(payload)
        self.assertEqual(user.username, payload['preferred_username'])
        self.assertEqual(user.email, payload['email'])
        self.assertEqual(user.is_staff, is_staff)

    def test_authenticate_credentials_user_updates_default_attributes(self):
        """ Test whether the user model is being assigned default fields from the payload. """

        username = 'gcostanza'
        old_email = 'tbone@gmail.com'
        new_email = 'koko@gmail.com'

        user = factories.UserFactory(email=old_email, username=username, is_staff=False)
        self.assertEqual(user.email, old_email)
        self.assertFalse(user.is_staff)

        payload = {'username': username, 'email': new_email, 'is_staff': True}

        user = JwtAuthentication().authenticate_credentials(payload)
        self.assertEqual(user.email, new_email)
        self.assertFalse(user.is_staff)

    @override_settings(
        EDX_DRF_EXTENSIONS={'JWT_PAYLOAD_USER_ATTRIBUTE_MAPPING': {'email': 'email', 'is_staff': 'is_staff'}}
    )
    def test_authenticate_credentials_user_attributes_custom_attributes(self):
        """ Test whether the user model is being assigned all custom fields from the payload. """

        username = 'ckramer'
        old_email = 'ckramer@hotmail.com'
        new_email = 'cosmo@hotmail.com'

        user = factories.UserFactory(email=old_email, username=username, is_staff=False)
        self.assertEqual(user.email, old_email)
        self.assertFalse(user.is_staff)

        payload = {'username': username, 'email': new_email, 'is_staff': True}

        user = JwtAuthentication().authenticate_credentials(payload)
        self.assertEqual(user.email, new_email)
        self.assertTrue(user.is_staff)

    @override_settings(
        EDX_DRF_EXTENSIONS={
            'JWT_PAYLOAD_USER_ATTRIBUTE_MAPPING': {
                'email': 'email',
                'is_staff': 'is_staff',
                'tags': 'tags',
                'fun_attr': 'fun_attr',
                'fruit': 'fruit'
            },
            'JWT_PAYLOAD_MERGEABLE_USER_ATTRIBUTES': [
                'tags',
                'fun_attr',
                'fruit'
            ]
        }
    )
    def test_authenticate_credentials_user_attributes_merge_attributes(self):
        """ Test whether the user model is being assigned all custom fields from the payload. """

        username = 'ckramer'
        email = 'ckramer@hotmail.com'
        old_tags = {'country': 'USA', 'browser': 'Firefox'}
        new_tags = {'browser': 'Chrome', 'new_attr': 'here!'}
        new_fun_attr = {'shiny': 'object'}
        expected_tags = {'country': 'USA', 'browser': 'Chrome', 'new_attr': 'here!'}
        old_fruit = {'fruit': 'apple'}

        user = factories.UserFactory(email=email, username=username, is_staff=False)
        setattr(user, 'tags', old_tags)
        setattr(user, 'fruit', old_fruit)
        self.assertEqual(user.email, email)
        self.assertFalse(user.is_staff)
        self.assertEqual(user.tags, old_tags)
        self.assertEqual(user.fruit, old_fruit)  # pylint: disable=no-member

        payload = {'username': username, 'email': email, 'is_staff': True, 'tags': new_tags, 'fun_attr': new_fun_attr}

        # Patch get_or_create so that our tags attribute is on the user object
        with mock.patch('edx_rest_framework_extensions.auth.jwt.authentication.get_user_model') as mock_get_user_model:
            mock_get_user_model().objects.get_or_create.return_value = (user, False)

            user = JwtAuthentication().authenticate_credentials(payload)
        self.assertEqual(user.tags, expected_tags)
        self.assertEqual(user.email, email)
        self.assertTrue(user.is_staff)
        self.assertEqual(user.fun_attr, new_fun_attr)
        self.assertEqual(user.fruit, old_fruit)

    @override_settings(
        EDX_DRF_EXTENSIONS={
            'JWT_PAYLOAD_USER_ATTRIBUTE_MAPPING': {'email': 'email', 'is_staff': 'is_staff', 'tags': 'tags'},
            'JWT_PAYLOAD_MERGEABLE_USER_ATTRIBUTES': ['tags']
        }
    )
    def test_authenticate_credentials_user_attributes_new_mergeable_attributes(self):
        """ Test whether the user model is being assigned all custom fields from the payload. """

        username = 'ckramer'
        email = 'ckramer@hotmail.com'
        new_tags = {'browser': 'Chrome'}

        user = factories.UserFactory(email=email, username=username, is_staff=False)
        self.assertEqual(user.email, email)
        self.assertFalse(user.is_staff)

        payload = {'username': username, 'email': email, 'is_staff': True, 'tags': new_tags}

        # Patch get_or_create so that our tags attribute is on the user object
        with mock.patch('edx_rest_framework_extensions.auth.jwt.authentication.get_user_model') as mock_get_user_model:
            mock_get_user_model().objects.get_or_create.return_value = (user, False)

            user = JwtAuthentication().authenticate_credentials(payload)
        self.assertEqual(user.tags, new_tags)
        self.assertEqual(user.email, email)
        self.assertTrue(user.is_staff)

    def test_authenticate_credentials_user_retrieval_failed(self):
        """ Verify exceptions raised during user retrieval are properly logged. """

        with mock.patch.object(User.objects, 'get_or_create', side_effect=ValueError):
            with mock.patch.object(Logger, 'exception') as logger:
                self.assertRaises(
                    AuthenticationFailed,
                    JwtAuthentication().authenticate_credentials,
                    {'username': 'test', 'email': 'test@example.com'}
                )
                logger.assert_called_with('[edx-drf-extensions] User retrieval failed for username test.')

    def test_authenticate_credentials_no_usernames(self):
        """ Verify an AuthenticationFailed exception is raised if the payload contains no username claim. """
        with self.assertRaises(AuthenticationFailed):
            JwtAuthentication().authenticate_credentials({'email': 'test@example.com'})

    @mock.patch.object(JwtAuthentication, 'enforce_csrf')
    @mock.patch('edx_rest_framework_extensions.auth.jwt.authentication.set_custom_attribute')
    def test_authenticate_with_correct_jwt_cookie(self, mock_set_custom_attribute, mock_enforce_csrf):
        """ Verify authenticate succeeds with a valid JWT cookie. """
        request = RequestFactory().post('/')
        request.COOKIES[jwt_cookie_name()] = self._get_test_jwt_token()
        drf_request = Request(request)

        assert JwtAuthentication().authenticate(drf_request)

        mock_enforce_csrf.assert_called_with(drf_request)
        mock_set_custom_attribute.assert_any_call('jwt_auth_result', 'success-cookie')
        set_custom_attribute_keys = [call.args[0] for call in mock_set_custom_attribute.call_args_list]
        assert 'jwt_auth_with_django_request' not in set_custom_attribute_keys

    @mock.patch.object(JwtAuthentication, 'enforce_csrf')
    @mock.patch('edx_rest_framework_extensions.auth.jwt.authentication.set_custom_attribute')
    def test_authenticate_with_correct_jwt_cookie_and_django_request(
        self, mock_set_custom_attribute, mock_enforce_csrf
    ):
        """
        Verify authenticate succeeds with a valid JWT cookie and a Django request.

        Note that JwtAuthentication is a DRF class, so a DRF request is expected. However,
        there is custom authentication code that passes in a Django request, so this test
        ensures backward compatibility. A custom attribute has been added to track down this
        custom authentication code.
        """
        request = RequestFactory().post('/')
        request.COOKIES[jwt_cookie_name()] = self._get_test_jwt_token()

        assert JwtAuthentication().authenticate(request)
        mock_enforce_csrf.assert_called_with(request)
        mock_set_custom_attribute.assert_any_call('jwt_auth_with_django_request', True)
        mock_set_custom_attribute.assert_any_call('jwt_auth_result', 'success-cookie')

    @mock.patch('edx_rest_framework_extensions.auth.jwt.authentication.set_custom_attribute')
    def test_authenticate_csrf_protected(self, mock_set_custom_attribute):
        """
        Ensure authenticate for JWTs properly handles CSRF errors.

        Note: With forgiving JWTs, all JWT cookie exceptions, including CSRF, will
        result in a None so that other authentication classes will also be checked.
        """
        request = RequestFactory().post('/')
        # Set a sample JWT cookie. We mock the auth response but we still want
        # to ensure that there is jwt set because there is other logic that
        # checks for the jwt to be set before moving forward with CSRF checks.
        request.COOKIES[jwt_cookie_name()] = 'foo'
        drf_request = Request(request)

        with mock.patch.object(JSONWebTokenAuthentication, 'authenticate', return_value=('mock-user', "mock-auth")):
            assert JwtAuthentication().authenticate(drf_request) is None
            mock_set_custom_attribute.assert_any_call('jwt_auth_result', 'forgiven-failure')

        mock_set_custom_attribute.assert_any_call(
            'jwt_auth_failed',
            "Exception:PermissionDenied('CSRF Failed: CSRF cookie not set.')",
        )

    @ddt.data(True, False)
    def test_get_decoded_jwt_from_auth(self, is_jwt_authentication):
        """ Verify get_decoded_jwt_from_auth returns the appropriate value. """

        # Mock out the `is_jwt_authenticated` method
        authentication.is_jwt_authenticated = lambda request: is_jwt_authentication

        jwt_token = self._get_test_jwt_token()
        mock_request_with_cookie = mock.Mock(COOKIES={}, auth=jwt_token)

        expected_decoded_jwt = jwt_decode_handler(jwt_token) if is_jwt_authentication else None

        decoded_jwt = authentication.get_decoded_jwt_from_auth(mock_request_with_cookie)
        self.assertEqual(expected_decoded_jwt, decoded_jwt)

    @mock.patch('edx_rest_framework_extensions.auth.jwt.authentication.set_custom_attribute')
    def test_authenticate_with_correct_jwt_authorization(self, mock_set_custom_attribute):
        """
        With JWT header it continues and validates the credentials.

        Note: CSRF protection should be skipped for this case, with no PermissionDenied.
        """
        jwt_token = self._get_test_jwt_token()
        request = RequestFactory().get('/', HTTP_AUTHORIZATION=f"JWT {jwt_token}")
        assert JwtAuthentication().authenticate(request)
        mock_set_custom_attribute.assert_any_call('jwt_auth_result', 'success-auth-header')

    @mock.patch('edx_rest_framework_extensions.auth.jwt.authentication.set_custom_attribute')
    def test_authenticate_with_incorrect_jwt_authorization(self, mock_set_custom_attribute):
        """ With JWT header it continues and validates the credentials and throws error. """
        auth_header = '{token_name} {token}'.format(token_name='JWT', token='wrongvalue')
        request = RequestFactory().get('/', HTTP_AUTHORIZATION=auth_header)
        with self.assertRaises(AuthenticationFailed):
            JwtAuthentication().authenticate(request)
        mock_set_custom_attribute.assert_any_call('jwt_auth_result', 'failed-auth-header')

    @mock.patch('edx_rest_framework_extensions.auth.jwt.authentication.set_custom_attribute')
    def test_authenticate_with_correct_jwt_authorization_and_bad_cookie(self, mock_set_custom_attribute):
        """
        With JWT header it continues and validates the credentials and ignores the invalid cookie.

        Note: CSRF protection should be skipped for this case, with no PermissionDenied.
        """
        jwt_token = self._get_test_jwt_token()
        request = RequestFactory().get('/', HTTP_AUTHORIZATION=f"JWT {jwt_token}")
        request.COOKIES[jwt_cookie_name()] = 'foo'
        assert JwtAuthentication().authenticate(request)
        mock_set_custom_attribute.assert_any_call('jwt_auth_result', 'success-auth-header')

    @mock.patch('edx_rest_framework_extensions.auth.jwt.authentication.set_custom_attribute')
    def test_authenticate_with_bearer_token(self, mock_set_custom_attribute):
        """ Returns a None for bearer header request. """
        auth_header = '{token_name} {token}'.format(token_name='Bearer', token='abc123')
        request = RequestFactory().get('/', HTTP_AUTHORIZATION=auth_header)
        self.assertIsNone(JwtAuthentication().authenticate(request))
        mock_set_custom_attribute.assert_any_call('jwt_auth_result', 'n/a')

    @override_settings(
        MIDDLEWARE=(
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
        ),
        ROOT_URLCONF='edx_rest_framework_extensions.auth.jwt.tests.test_authentication',
    )
    @mock.patch('edx_rest_framework_extensions.auth.jwt.authentication.set_custom_attribute')
    def test_authenticate_jwt_and_session_mismatch(self, mock_set_custom_attribute):
        """ Tests monitoring for JWT cookie when there is a session user mismatch """
        session_user = factories.UserFactory(id=111, username='session-name')
        jwt_user = factories.UserFactory(id=222, username='jwt-name')
        self.client.cookies = SimpleCookie({
            jwt_cookie_name(): self._get_test_jwt_token(user=jwt_user),
        })

        self.client.force_login(session_user)
        response = self.client.get(reverse('authenticated-view'))

        mock_set_custom_attribute.assert_any_call('jwt_auth_result', 'success-cookie')
        mock_set_custom_attribute.assert_any_call(
            'jwt_cookie_lms_user_id', jwt_user.id  # pylint: disable=no-member
        )
        mock_set_custom_attribute.assert_any_call('jwt_auth_mismatch_session_username', session_user.username)
        mock_set_custom_attribute.assert_any_call('jwt_auth_mismatch_jwt_cookie_username', jwt_user.username)
        assert response.status_code == 200

    @override_settings(
        MIDDLEWARE=(
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
        ),
        ROOT_URLCONF='edx_rest_framework_extensions.auth.jwt.tests.test_authentication',
    )
    @mock.patch('edx_rest_framework_extensions.auth.jwt.authentication.set_custom_attribute')
    def test_authenticate_jwt_and_session_mismatch_bad_signature_cookie(self, mock_set_custom_attribute):
        """ Tests monitoring for JWT cookie with a bad signature when there is a session user mismatch """
        session_user = factories.UserFactory(id=111, username='session-name')
        jwt_user = factories.UserFactory(id=222, username='jwt-name')
        self.client.cookies = SimpleCookie({
            jwt_cookie_name(): self._get_test_jwt_token(user=jwt_user, is_valid_signature=False),
        })

        self.client.force_login(session_user)
        response = self.client.get(reverse('authenticated-view'))

        mock_set_custom_attribute.assert_any_call(
            'jwt_cookie_lms_user_id', jwt_user.id  # pylint: disable=no-member
        )
        mock_set_custom_attribute.assert_any_call('jwt_auth_result', 'user-mismatch-failure')
        mock_set_custom_attribute.assert_any_call('jwt_auth_mismatch_session_username', session_user.username)
        mock_set_custom_attribute.assert_any_call('jwt_auth_mismatch_jwt_cookie_username', jwt_user.username)
        assert response.status_code == 401

    @override_settings(
        MIDDLEWARE=(
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
        ),
        ROOT_URLCONF='edx_rest_framework_extensions.auth.jwt.tests.test_authentication',
    )
    @mock.patch('edx_rest_framework_extensions.auth.jwt.authentication.set_custom_attribute')
    def test_authenticate_jwt_and_session_mismatch_invalid_cookie(self, mock_set_custom_attribute):
        """ Tests monitoring for invalid JWT cookie when there is a session user mismatch """
        session_user = factories.UserFactory(id=111, username='session-name')
        self.client.cookies = SimpleCookie({
            jwt_cookie_name(): 'invalid-cookie',
        })

        self.client.force_login(session_user)
        response = self.client.get(reverse('authenticated-view'))

        mock_set_custom_attribute.assert_any_call('jwt_cookie_lms_user_id', None)
        mock_set_custom_attribute.assert_any_call('jwt_cookie_unsafe_decode_issue', 'decode-error')
        mock_set_custom_attribute.assert_any_call('jwt_auth_result', 'user-mismatch-failure')
        mock_set_custom_attribute.assert_any_call('jwt_auth_mismatch_session_username', session_user.username)
        mock_set_custom_attribute.assert_any_call('jwt_auth_mismatch_jwt_cookie_username', None)
        assert response.status_code == 401

    @override_settings(
        MIDDLEWARE=(
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
        ),
        ROOT_URLCONF='edx_rest_framework_extensions.auth.jwt.tests.test_authentication',
    )
    @mock.patch('edx_rest_framework_extensions.auth.jwt.authentication.set_custom_attribute')
    def test_authenticate_jwt_and_session_match(self, mock_set_custom_attribute):
        """ Tests monitoring for JWT cookie when session user matches """
        test_user = factories.UserFactory()
        self.client.cookies = SimpleCookie({
            jwt_cookie_name(): self._get_test_jwt_token(user=test_user),
        })

        self.client.force_login(test_user)
        response = self.client.get(reverse('authenticated-view'))

        set_custom_attribute_keys = [call.args[0] for call in mock_set_custom_attribute.call_args_list]
        assert 'jwt_auth_mismatch_session_username' not in set_custom_attribute_keys
        assert 'jwt_auth_mismatch_jwt_cookie_username' not in set_custom_attribute_keys
        assert response.status_code == 200

    @override_settings(
        MIDDLEWARE=(
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
        ),
        ROOT_URLCONF='edx_rest_framework_extensions.auth.jwt.tests.test_authentication',
    )
    @mock.patch('edx_rest_framework_extensions.auth.jwt.authentication.set_custom_attribute')
    def test_authenticate_jwt_and_no_session(self, mock_set_custom_attribute):
        """ Tests monitoring for JWT cookie when there is no session """
        test_user = factories.UserFactory()
        self.client.cookies = SimpleCookie({
            jwt_cookie_name(): self._get_test_jwt_token(user=test_user),
        })

        # unlike other tests, there is no force_login call to start the session
        response = self.client.get(reverse('authenticated-view'))

        set_custom_attribute_keys = [call.args[0] for call in mock_set_custom_attribute.call_args_list]
        assert 'jwt_auth_mismatch_session_username' not in set_custom_attribute_keys
        assert 'jwt_auth_mismatch_jwt_cookie_username' not in set_custom_attribute_keys
        assert response.status_code == 200

    @override_settings(
        EDX_DRF_EXTENSIONS={
            ENABLE_SET_REQUEST_USER_FOR_JWT_COOKIE: True,
        },
        MIDDLEWARE=(
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
            'edx_rest_framework_extensions.auth.jwt.middleware.JwtAuthCookieMiddleware',
        ),
        ROOT_URLCONF='edx_rest_framework_extensions.auth.jwt.tests.test_authentication',
    )
    @mock.patch('edx_rest_framework_extensions.auth.jwt.authentication.set_custom_attribute')
    def test_authenticate_mismatch_with_set_request_user(self, mock_set_custom_attribute):
        """
        Tests failure for JWT cookie when there is a session user mismatch with a request to set user.

        - This tests coordination between ENABLE_SET_REQUEST_USER_FOR_JWT_COOKIE in middleware and JwtAuthentication.
        - This test is kept with the rest of the JWT vs session user tests.
        """
        session_user = factories.UserFactory(id=111, username='session-name')
        jwt_user = factories.UserFactory(id=222, username='jwt-name')
        jwt_header_payload, jwt_signature = self._get_test_jwt_token_payload_and_signature(user=jwt_user)
        # Cookie parts will be recombined by JwtAuthCookieMiddleware
        self.client.cookies = SimpleCookie({
            jwt_cookie_header_payload_name(): jwt_header_payload,
            jwt_cookie_signature_name(): jwt_signature,
        })

        self.client.force_login(session_user)

        with self.assertRaises(JwtSessionUserMismatchError):
            response = self.client.get(reverse('authenticated-view'))
            assert response.status_code == 401

        mock_set_custom_attribute.assert_any_call('jwt_auth_mismatch_session_username', session_user.username)
        mock_set_custom_attribute.assert_any_call('jwt_auth_mismatch_jwt_cookie_username', jwt_user.username)
        mock_set_custom_attribute.assert_any_call(
            'jwt_cookie_lms_user_id', jwt_user.id  # pylint: disable=no-member
        )
        mock_set_custom_attribute.assert_any_call('jwt_auth_result', 'user-mismatch-enforced-failure')
        mock_set_custom_attribute.assert_any_call('jwt_auth_failed', mock.ANY)

    @override_settings(
        EDX_DRF_EXTENSIONS={
            ENABLE_SET_REQUEST_USER_FOR_JWT_COOKIE: True,
        },
        MIDDLEWARE=(
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
            'edx_rest_framework_extensions.auth.jwt.middleware.JwtAuthCookieMiddleware',
        ),
        ROOT_URLCONF='edx_rest_framework_extensions.auth.jwt.tests.test_authentication',
    )
    @mock.patch('edx_rest_framework_extensions.auth.jwt.authentication.set_custom_attribute')
    def test_authenticate_jwt_and_no_session_and_set_request_user(self, mock_set_custom_attribute):
        """
        Tests success for JWT cookie when there is no session user and there is a request to set user.

        - This tests coordination between ENABLE_SET_REQUEST_USER_FOR_JWT_COOKIE in middleware and JwtAuthentication.
        - This test is kept with the rest of the JWT vs session user tests.
        """
        test_user = factories.UserFactory()
        jwt_lms_user_id = 222
        jwt_header_payload, jwt_signature = self._get_test_jwt_token_payload_and_signature(
            user=test_user, lms_user_id=jwt_lms_user_id
        )
        # Cookie parts will be recombined by JwtAuthCookieMiddleware
        self.client.cookies = SimpleCookie({
            jwt_cookie_header_payload_name(): jwt_header_payload,
            jwt_cookie_signature_name(): jwt_signature,
        })

        # unlike other tests, there is no force_login call to start the session
        response = self.client.get(reverse('authenticated-view'))

        mock_set_custom_attribute.assert_any_call('jwt_cookie_lms_user_id', jwt_lms_user_id)
        set_custom_attribute_keys = [call.args[0] for call in mock_set_custom_attribute.call_args_list]
        assert 'jwt_auth_mismatch_session_username' not in set_custom_attribute_keys
        assert 'jwt_auth_mismatch_jwt_cookie_username' not in set_custom_attribute_keys
        assert response.status_code == 200

    def _get_test_jwt_token(self, user=None, is_valid_signature=True, lms_user_id=None):
        """ Returns a test jwt token for the provided user """
        test_user = factories.UserFactory() if user is None else user
        payload = generate_latest_version_payload(test_user)
        if lms_user_id:
            # In other services, the LMS user id in the JWT would not be the user's id.
            payload['user_id'] = lms_user_id
        if is_valid_signature:
            jwt_token = generate_jwt_token(payload)
        else:
            jwt_token = generate_jwt_token(payload, signing_key='invalid-key')
        return jwt_token

    def _get_test_jwt_token_payload_and_signature(self, user=None, lms_user_id=None):
        """ Returns a test jwt token split into payload and signature """
        jwt_token = self._get_test_jwt_token(user=user, lms_user_id=lms_user_id)
        jwt_token_parts = jwt_token.split('.')
        header_and_payload = '.'.join(jwt_token_parts[0:2])
        signature = jwt_token_parts[2]
        return header_and_payload, signature


class TestLowestJWTException:
    """
    Test that we're getting the correct exception out of a stack of exceptions when checking a JWT for auth Fails.

    The exception closest to us does not have sufficient useful information so we have to see what other exceptions the
    current exception came from.
    """
    # pylint: disable=broad-exception-caught, raise-missing-from, unused-variable, protected-access

    def test_jwt_exception_in_the_middle(self):
        mock_jwt_exception = jwt_exceptions.DecodeError("Not enough segments")
        try:
            try:
                try:
                    raise Exception("foo")
                except Exception as exception:
                    raise mock_jwt_exception
            except Exception as exception:
                raise AuthenticationFailed()
        except Exception as exception:
            e = authentication._deepest_jwt_exception(exception)
            assert e == mock_jwt_exception

    def test_jwt_exception_at_the_bottom(self):
        mock_jwt_exception = jwt_exceptions.DecodeError("Not enough segments")
        try:
            try:
                try:
                    raise mock_jwt_exception
                except Exception as exception:
                    raise Exception("foo")
            except Exception as exception:
                raise AuthenticationFailed()
        except Exception as exception:
            e = authentication._deepest_jwt_exception(exception)
            assert e == mock_jwt_exception

    def test_jwt_exception_at_the_top(self):
        mock_jwt_exception = jwt_exceptions.DecodeError("Not enough segments")
        try:
            try:
                try:
                    raise Exception("foo")
                except Exception as exception:
                    raise AuthenticationFailed()
            except Exception as exception:
                raise mock_jwt_exception
        except Exception as exception:
            e = authentication._deepest_jwt_exception(exception)
            assert e == mock_jwt_exception

    def test_multiple_jwt_exceptions(self):
        mock_jwt_exception = jwt_exceptions.DecodeError("Not enough segments")
        try:
            try:
                try:
                    raise Exception("foo")
                except Exception as exception:
                    raise mock_jwt_exception
            except Exception as exception:
                raise jwt_exceptions.InvalidTokenError()
        except Exception as exception:
            e = authentication._deepest_jwt_exception(exception)
            assert e == mock_jwt_exception



""" Tests for utility functions. """
import copy
from unittest import mock

import ddt
import jwt
from django.conf import settings
from django.test import TestCase, override_settings

from edx_rest_framework_extensions.auth.jwt.decoder import (
    decode_jwt_filters,
    decode_jwt_is_restricted,
    decode_jwt_scopes,
    get_asymmetric_only_jwt_decode_handler,
    jwt_decode_handler,
    unsafe_jwt_decode_handler,
)
from edx_rest_framework_extensions.auth.jwt.tests.utils import (
    generate_asymmetric_jwt_token,
    generate_jwt_token,
    generate_latest_version_payload,
    generate_unversioned_payload,
)
from edx_rest_framework_extensions.tests.factories import UserFactory


def exclude_from_jwt_auth_setting(key):
    """
    Clone the JWT_AUTH setting dict and remove the given key.
    """
    jwt_auth = copy.deepcopy(settings.JWT_AUTH)
    del jwt_auth[key]
    return jwt_auth


def update_jwt_auth_setting(jwt_auth_overrides):
    """
    Clone the JWT_AUTH setting dict and update it with the given overrides.
    """
    jwt_auth = copy.deepcopy(settings.JWT_AUTH)
    jwt_auth.update(jwt_auth_overrides)
    return jwt_auth


@ddt.ddt
class JWTDecodeHandlerTests(TestCase):
    """ Tests for the `jwt_decode_handler` utility function. """
    def setUp(self):
        super().setUp()
        self.user = UserFactory()
        self.payload = generate_latest_version_payload(self.user)
        self.jwt = generate_jwt_token(self.payload)

    def test_success(self):
        """
        Confirms that the format of the valid response from the token decoder matches the payload
        """
        self.assertDictEqual(jwt_decode_handler(self.jwt), self.payload)

    @ddt.data(*settings.JWT_AUTH['JWT_ISSUERS'])
    def test_valid_token_multiple_valid_issuers(self, jwt_issuer):
        """
        Validates that a valid token is properly decoded given a list of multiple valid issuers
        """

        # Verify that each valid issuer is properly matched against the valid issuers list
        # and used to decode the token that was generated using said valid issuer data
        self.payload['iss'] = jwt_issuer['ISSUER']
        token = generate_jwt_token(self.payload, jwt_issuer['SECRET_KEY'])
        self.assertEqual(jwt_decode_handler(token), self.payload)

    def test_failure_invalid_issuer(self):
        """
        Verifies the function logs decode failures with invalid issuer,
        and raises an InvalidTokenError if the token cannot be decoded
        """

        # Create tokens using each invalid issuer and attempt to decode them against
        # the valid issuers list, which won't work
        with mock.patch('edx_rest_framework_extensions.auth.jwt.decoder.logger') as patched_log:
            with self.assertRaises(jwt.InvalidTokenError):
                self.payload['iss'] = 'invalid-issuer'
                # signing key of None will use the default valid signing key
                valid_signing_key = None
                # Generate a token using the invalid issuer data
                token = generate_jwt_token(self.payload, valid_signing_key)
                # Attempt to decode the token against the entries in the valid issuers list,
                # which will fail with an InvalidTokenError
                jwt_decode_handler(token)

            msg = "Token decode failed due to mismatched issuer [%s]"
            patched_log.info.assert_any_call(msg, 'invalid-issuer')

    def test_failure_invalid_token(self):
        """
        Verifies the function logs decode failures, and raises an InvalidTokenError if the token cannot be decoded
        """

        # Create tokens using each invalid issuer and attempt to decode them against
        # the valid issuers list, which won't work
        with mock.patch('edx_rest_framework_extensions.auth.jwt.decoder.logger') as patched_log:
            with self.assertRaises(jwt.InvalidTokenError):
                # Attempt to decode an invalid token, which will fail with an InvalidTokenError
                jwt_decode_handler("invalid.token")

            patched_log.exception.assert_any_call("Token verification failed.")

    @override_settings(JWT_AUTH=exclude_from_jwt_auth_setting('JWT_SUPPORTED_VERSION'))
    def test_supported_jwt_version_not_specified(self):
        """
        Verifies the JWT is decoded successfully when the JWT_SUPPORTED_VERSION setting is not specified.
        """
        token = generate_jwt_token(self.payload)
        self.assertDictEqual(jwt_decode_handler(token), self.payload)

    @ddt.data(None, '0.5.0', '1.0.0', '1.0.5', '1.5.0', '1.5.5')
    def test_supported_jwt_version(self, jwt_version):
        """
        Verifies the JWT is decoded successfully with different supported versions in the token.
        """
        jwt_payload = generate_latest_version_payload(self.user, version=jwt_version)
        token = generate_jwt_token(jwt_payload)
        self.assertDictEqual(jwt_decode_handler(token), jwt_payload)

    @override_settings(JWT_AUTH=update_jwt_auth_setting({'JWT_SUPPORTED_VERSION': '0.5.0'}))
    def test_unsupported_jwt_version(self):
        """
        Verifies the function logs decode failures, and raises an
        InvalidTokenError if the token version is not supported.
        """
        with mock.patch('edx_rest_framework_extensions.auth.jwt.decoder.logger') as patched_log:
            with self.assertRaises(jwt.InvalidTokenError):
                token = generate_jwt_token(self.payload)
                jwt_decode_handler(token)

            msg = "Token decode failed due to unsupported JWT version number [%s]"
            patched_log.info.assert_any_call(msg, '1.1.0')

    def test_upgrade(self):
        """
        Verifies the JWT is upgraded when an old (starting) version is provided.
        """
        jwt_payload = generate_unversioned_payload(self.user)
        token = generate_jwt_token(jwt_payload)

        upgraded_payload = generate_latest_version_payload(self.user, version='1.0.0')

        # Keep time-related values constant for full-proof comparison.
        upgraded_payload['iat'], upgraded_payload['exp'] = jwt_payload['iat'], jwt_payload['exp']
        self.assertDictEqual(jwt_decode_handler(token), upgraded_payload)

    def test_failure_invalid_signature(self):
        """
        Verifies the function logs decode failures with invalid signature,
        and raises an InvalidTokenError if the token cannot be decoded
        """
        # Create tokens using each invalid signature and attempt to decode them against
        # the valid signature.
        with mock.patch('edx_rest_framework_extensions.auth.jwt.decoder.logger') as patched_log:
            with self.assertRaises(jwt.InvalidTokenError):
                invalid_signing_key = 'invalid-secret-key'

                # Generate a token using the invalid signing key data
                token = generate_jwt_token(self.payload, invalid_signing_key)
                # Attempt to decode the token against invalid signature,
                # which will fail with an InvalidTokenError
                jwt_decode_handler(token)

            patched_log.exception.assert_any_call("Token verification failed.")

    @ddt.data("exp", "iat")
    def test_required_claims(self, claim):
        """
        Verify that tokens that do not carry 'exp' or 'iat' claims are rejected
        """
        # Deletes required claim from payload
        del self.payload[claim]
        token = generate_jwt_token(self.payload)
        with self.assertRaises(jwt.MissingRequiredClaimError):
            # Decode to see if MissingRequiredClaimError exception is raised or not
            jwt_decode_handler(token)

    def test_failure_decode_symmetric_set_as_False(self):
        """
        Verifies the function logs decode failures with symmetric token set as false,
        and raises an InvalidTokenError if token is symmetric
        """
        # Create valid token symmetric jwt token and set decode_symmetric_token as False.
        with mock.patch('edx_rest_framework_extensions.auth.jwt.decoder.logger') as patched_log:
            with self.assertRaises(jwt.InvalidTokenError):
                token = generate_jwt_token(self.payload)
                jwt_decode_handler(token, decode_symmetric_token=False)

            patched_log.exception.assert_any_call("Token verification failed.")

    def test_success_asymmetric_jwt_decode(self):
        """
        Validates that a valid asymmetric token is properly decoded
        """
        token = generate_asymmetric_jwt_token(self.payload)
        self.assertEqual(get_asymmetric_only_jwt_decode_handler(token), self.payload)

    @mock.patch('edx_rest_framework_extensions.auth.jwt.decoder.set_custom_attribute')
    def test_keyset_size_and_other_monitoring(self, mock_set_custom_attribute):
        """
        Validates a variety of custom attributes are recorded, including the keyset size.
        """
        asymmetric_token = generate_asymmetric_jwt_token(self.payload)
        symmetric_token = generate_jwt_token(self.payload)

        # The secret key is included by default making a list of length 2, but for
        # asymmetric-only there is only 1 key in the keyset.
        self.assertEqual(jwt_decode_handler(asymmetric_token), self.payload)
        self.assertEqual(get_asymmetric_only_jwt_decode_handler(asymmetric_token), self.payload)
        self.assertEqual(jwt_decode_handler(symmetric_token), self.payload)

        assert mock_set_custom_attribute.call_args_list == [
            mock.call('jwt_auth_check_symmetric_key', True),
            mock.call('jwt_auth_verify_asymmetric_keys_count', 1),
            mock.call('jwt_auth_asymmetric_verified', True),
            mock.call('jwt_auth_issuer', 'test-issuer-1'),
            mock.call('jwt_auth_issuer_verification', 'matches-first-issuer'),

            mock.call('jwt_auth_check_symmetric_key', False),
            mock.call('jwt_auth_verify_asymmetric_keys_count', 1),
            mock.call('jwt_auth_asymmetric_verified', True),
            mock.call('jwt_auth_issuer', 'test-issuer-1'),
            mock.call('jwt_auth_issuer_verification', 'matches-first-issuer'),

            mock.call('jwt_auth_check_symmetric_key', True),
            mock.call('jwt_auth_verify_asymmetric_keys_count', 1),
            mock.call('jwt_auth_verify_all_keys_count', 2),
            mock.call('jwt_auth_symmetric_verified', True),
            mock.call('jwt_auth_issuer', 'test-issuer-1'),
            mock.call('jwt_auth_issuer_verification', 'matches-first-issuer'),
        ]

    def test_unsafe_success_with_invalid_token(self):
        """
        Verifies unsafe decode is successful, even with invalid claims and signature
        """
        self.payload['iss'] = 'invalid-iss'
        self.payload['exp'] = 'invalid-exp'
        self.payload['aud'] = 'invalid-aud'
        invalid_signing_key = 'invalid-secret-key'

        # Generate a token using the invalid signing key
        token = generate_jwt_token(self.payload, invalid_signing_key)
        decoded_token = unsafe_jwt_decode_handler(token)
        assert decoded_token['preferred_username'] is not None


def _jwt_decode_handler_with_defaults(token):  # pylint: disable=unused-argument
    """
    Accepts anything as a token and returns a fake JWT payload with defaults.
    """
    return {
        'scopes': ['fake:scope'],
        'is_restricted': True,
        'filters': ['fake:filter'],
    }


def _jwt_decode_handler_no_defaults(token):  # pylint: disable=unused-argument
    """
    Accepts anything as a token and returns a fake JWT payload with no defaults.
    """
    return {}


@ddt.ddt
class JWTDecodeHandlerSettingTests(TestCase):
    """
    Tests to ensure utility functions respect JWT_DECODE_HANDLER setting.

    Note: An attempt was made to use ``override_settings`` to actually set
    ``JWT_DECODE_HANDLER``, but clean-up of the tests in tearDown was not working,
    even after reloading the module, and it was failing other tests in the test suite.
    """
    NORMALLY_INVALID_TOKEN = 'this is a valid jwt only with fake_jwt_decode_handler'

    @ddt.data(
        ('_jwt_decode_handler_with_defaults', ['fake:scope']),
        ('_jwt_decode_handler_no_defaults', [])
    )
    @ddt.unpack
    @mock.patch('edx_rest_framework_extensions.auth.jwt.decoder.api_settings')
    def test_decode_jwt_scopes(self, jwt_decode_handler_name, expected_scope, mock_api_settings):
        mock_api_settings.JWT_DECODE_HANDLER = globals()[jwt_decode_handler_name]
        scopes = decode_jwt_scopes(self.NORMALLY_INVALID_TOKEN)
        self.assertEqual(scopes, expected_scope)

    @ddt.data(
        ('_jwt_decode_handler_with_defaults', True),
        ('_jwt_decode_handler_no_defaults', False)
    )
    @ddt.unpack
    @mock.patch('edx_rest_framework_extensions.auth.jwt.decoder.api_settings')
    def test_decode_jwt_is_restricted(self, jwt_decode_handler_name, expected_is_restricted, mock_api_settings):
        mock_api_settings.JWT_DECODE_HANDLER = globals()[jwt_decode_handler_name]
        is_restricted = decode_jwt_is_restricted(self.NORMALLY_INVALID_TOKEN)
        self.assertEqual(is_restricted, expected_is_restricted)

    @ddt.data(
        ('_jwt_decode_handler_with_defaults', [['fake', 'filter']]),
        ('_jwt_decode_handler_no_defaults', [])
    )
    @ddt.unpack
    @mock.patch('edx_rest_framework_extensions.auth.jwt.decoder.api_settings')
    def test_decode_jwt_filters(self, jwt_decode_handler_name, expected_filter, mock_api_settings):
        mock_api_settings.JWT_DECODE_HANDLER = globals()[jwt_decode_handler_name]
        filters = decode_jwt_filters(self.NORMALLY_INVALID_TOKEN)
        self.assertEqual(filters, expected_filter)



"""
Application configuration constants and code.
"""

# .. toggle_name: EDX_DRF_EXTENSIONS[ENABLE_SET_REQUEST_USER_FOR_JWT_COOKIE]
# .. toggle_implementation: DjangoSetting
# .. toggle_default: False
# .. toggle_description: Toggle for setting request.user with jwt cookie authentication. This makes the JWT cookie
#      user available to middleware while processing the request, if the session user wasn't already available. This
#      requires JwtAuthCookieMiddleware to work.
# .. toggle_use_cases: temporary
# .. toggle_creation_date: 2019-10-15
# .. toggle_target_removal_date: 2024-12-31
# .. toggle_warning: This feature caused a memory leak in edx-platform. This toggle is temporary only if we can make it
#      work in all services, or find a replacement. Consider making this a permanent toggle instead.
# .. toggle_tickets: ARCH-1210, ARCH-1199, ARCH-1197
ENABLE_SET_REQUEST_USER_FOR_JWT_COOKIE = 'ENABLE_SET_REQUEST_USER_FOR_JWT_COOKIE'
